import { Path, PathArray } from './path';
import { Source } from './source';
export interface TypeMismatch {
    type: Schema<unknown>;
    value: unknown;
    path: PathArray;
}
export interface ValidateOptions {
    path?: PathArray;
    allowExtraFields?: boolean;
}
export interface RestrictOptions {
    fillEmpty?: boolean;
    fillZero?: boolean;
    coerce?: boolean;
}
export declare abstract class Schema<T> {
    abstract validate(value: unknown, options?: ValidateOptions): TypeMismatch[];
    abstract restrict(value: unknown, options?: RestrictOptions): T | undefined;
    abstract zeroValue(): T;
    abstract toSource(): Source;
    abstract child(index: string | number): Schema<unknown> | undefined;
    abstract isRecursive(): boolean;
    atPath(path: Path): Schema<unknown> | undefined;
    isScalar(): this is ScalarType<T>;
    isArray(): this is ArrayType<unknown>;
    isObject(): this is ObjectType | DictionaryType<unknown>;
}
export declare abstract class ScalarType<T> extends Schema<T> {
    child(): undefined;
    atPath(): undefined;
    isScalar(): this is ScalarType<T>;
    isRecursive(): boolean;
}
export declare class StringType extends ScalarType<string> {
    static readonly self: StringType;
    private constructor();
    toSource(): 'string';
    validate(value: unknown, { path }?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): string | undefined;
    zeroValue(): string;
}
export declare class FloatType extends ScalarType<number> {
    static readonly self: FloatType;
    private constructor();
    toSource(): 'float';
    validate(value: unknown, { path }?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): number | undefined;
    zeroValue(): number;
}
export declare class IntegerType extends ScalarType<number> {
    static readonly self: IntegerType;
    private constructor();
    toSource(): 'integer';
    validate(value: unknown, { path }?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): number | undefined;
    zeroValue(): number;
}
export declare class BinaryType extends ScalarType<Uint8Array> {
    static readonly self: BinaryType;
    private constructor();
    toSource(): 'binary';
    validate(value: unknown, { path }?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): Uint8Array | undefined;
    zeroValue(): Uint8Array;
}
export declare class DateType extends ScalarType<Date> {
    static readonly self: DateType;
    private constructor();
    toSource(): 'date';
    validate(value: unknown, { path }?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): Date | undefined;
    zeroValue(): Date;
}
export declare class BooleanType extends ScalarType<boolean> {
    static readonly self: BooleanType;
    private constructor();
    toSource(): 'boolean';
    validate(value: unknown, { path }?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): boolean | undefined;
    zeroValue(): boolean;
}
export declare class NullType extends ScalarType<null> {
    static readonly self: NullType;
    private constructor();
    toSource(): null;
    readonly source: null;
    validate(value: unknown, { path }?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): null | undefined;
    zeroValue(): null;
}
export declare class OneOfType extends ScalarType<string | number | boolean | null | Uint8Array | Date> {
    readonly subtypes: readonly ScalarType<string | number | boolean | null | Uint8Array | Date>[];
    constructor(subtypes: readonly ScalarType<string | number | boolean | null | Uint8Array | Date>[]);
    toSource(): Source;
    validate(value: unknown, options?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): string | number | boolean | null | Uint8Array | Date | undefined;
    zeroValue(): string | number | boolean | null | Uint8Array | Date;
}
export declare class EnumType extends ScalarType<string | number | boolean | null> {
    readonly members: readonly (string | number | boolean | null)[];
    constructor(members: readonly (string | number | boolean | null)[]);
    toSource(): Source;
    validate(value: unknown, { path }?: ValidateOptions): TypeMismatch[];
    restrict(value: unknown, options?: RestrictOptions): string | number | boolean | null | undefined;
    zeroValue(): string | number | boolean | null;
}
export declare class ArrayType<T> extends Schema<T[]> {
    readonly elementType: Schema<T>;
    constructor(elementType: Schema<T>);
    toSource(): Source;
    validate(value: unknown, options?: ValidateOptions): TypeMismatch[];
    child(path: string | number): Schema<T> | undefined;
    isArray(): this is ArrayType<T>;
    isRecursive(): boolean;
    restrict(value: unknown, options?: RestrictOptions): T[] | undefined;
    zeroValue(): T[];
}
export declare class DictionaryType<T> extends Schema<{
    [key: string]: T;
}> {
    readonly valueType: Schema<T>;
    constructor(valueType: Schema<T>);
    toSource(): Source;
    validate(value: unknown, options?: ValidateOptions): TypeMismatch[];
    child(path: string | number): Schema<T> | undefined;
    isObject(): this is DictionaryType<T>;
    isRecursive(): boolean;
    restrict(value: unknown, options?: RestrictOptions): {
        [key: string]: T;
    } | undefined;
    zeroValue(): {
        [key: string]: T;
    };
}
export declare class ObjectType extends Schema<object> {
    readonly entries: {
        readonly [key: string]: Schema<unknown>;
    };
    constructor(entries: {
        readonly [key: string]: Schema<unknown>;
    });
    toSource(): Source;
    validate(value: unknown, options?: ValidateOptions): TypeMismatch[];
    child(path: string | number): Schema<unknown> | undefined;
    isObject(): this is ObjectType;
    isRecursive(): boolean;
    restrict(value: unknown, options?: RestrictOptions): object | undefined;
    zeroValue(): object;
}
export declare function compileSchema(source: Source, path?: PathArray, scope?: {
    readonly [label: string]: Schema<unknown>;
}): Schema<unknown>;
//# sourceMappingURL=schema.d.ts.map