"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.compileSchema = compileSchema;
exports.ObjectType = exports.DictionaryType = exports.ArrayType = exports.EnumType = exports.OneOfType = exports.NullType = exports.BooleanType = exports.DateType = exports.BinaryType = exports.IntegerType = exports.FloatType = exports.StringType = exports.ScalarType = exports.Schema = void 0;

var _assert = _interopRequireDefault(require("assert"));

var _lodash = _interopRequireDefault(require("lodash.flatmap"));

var _lodash2 = _interopRequireDefault(require("lodash.frompairs"));

var _lodash3 = _interopRequireDefault(require("lodash.isplainobject"));

var _lodash4 = _interopRequireDefault(require("lodash.partition"));

var _errors = require("./errors");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function isObject(x) {
  return (0, _lodash3.default)(x);
}

class Schema {
  atPath(path) {
    // eslint-disable-next-line @typescript-eslint/no-this-alias
    let s = this;

    for (const i of path) {
      if (!s) {
        break;
      }

      s = s.child(i);
    }

    return s;
  }

  isScalar() {
    return false;
  }

  isArray() {
    return false;
  }

  isObject() {
    return false;
  }

}

exports.Schema = Schema;

class ScalarType extends Schema {
  child() {
    return undefined;
  }

  atPath() {
    return undefined;
  }

  isScalar() {
    return true;
  }

  isRecursive() {
    return false;
  }

}

exports.ScalarType = ScalarType;

class StringType extends ScalarType {
  static self = new StringType();

  constructor() {
    super();
  }

  toSource() {
    return 'string';
  }

  validate(value, {
    path = []
  } = {}) {
    return typeof value === 'string' ? [] : [{
      type: this,
      value,
      path
    }];
  }

  restrict(value, options = {}) {
    if (typeof value === 'string') {
      return value;
    }

    if (options.coerce) {
      if (typeof value === 'number' || typeof value === 'boolean' || value === null) {
        return `${value}`;
      }

      if (value instanceof Uint8Array) {
        return Buffer.from(value).toString('base64');
      }

      if (value instanceof Date) {
        return value.toISOString();
      }
    }

    if (options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return '';
  }

}

exports.StringType = StringType;

class FloatType extends ScalarType {
  static self = new FloatType();

  constructor() {
    super();
  }

  toSource() {
    return 'float';
  }

  validate(value, {
    path = []
  } = {}) {
    return typeof value === 'number' ? [] : [{
      type: this,
      value,
      path
    }];
  }

  restrict(value, options = {}) {
    if (typeof value === 'number') {
      return value;
    }

    if (options.coerce) {
      if (typeof value === 'string') {
        return +value;
      } else if (value === true) {
        return 1;
      } else if (value === false || value == null) {
        return 0;
      }
    }

    if (options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return 0;
  }

}

exports.FloatType = FloatType;

class IntegerType extends ScalarType {
  static self = new IntegerType();

  constructor() {
    super();
  }

  toSource() {
    return 'integer';
  }

  validate(value, {
    path = []
  } = {}) {
    return typeof value === 'number' && value === Math.floor(value) ? [] : [{
      type: this,
      value,
      path
    }];
  }

  restrict(value, options = {}) {
    if (typeof value === 'number' && value === Math.floor(value)) {
      return value;
    }

    if (options.coerce) {
      if (typeof value === 'number') {
        return Math.round(value);
      } else if (typeof value === 'string') {
        const parsed = parseInt(value);

        if (!isNaN(parsed)) {
          return parsed;
        }
      } else if (value === true) {
        return 1;
      } else if (value === false || value == null) {
        return 0;
      } else if (value instanceof Date) {
        return value.getTime();
      }
    }

    if (options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return 0;
  }

}

exports.IntegerType = IntegerType;

class BinaryType extends ScalarType {
  static self = new BinaryType();

  constructor() {
    super();
  }

  toSource() {
    return 'binary';
  }

  validate(value, {
    path = []
  } = {}) {
    return value instanceof Uint8Array ? [] : [{
      type: this,
      value,
      path
    }];
  }

  restrict(value, options = {}) {
    if (value instanceof Uint8Array) {
      return value;
    }

    if (options.coerce && typeof value === 'string') {
      try {
        return Buffer.from(value, 'base64');
      } catch (e) {// do nothing
      }
    }

    if (options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return new Uint8Array();
  }

}

exports.BinaryType = BinaryType;

class DateType extends ScalarType {
  static self = new DateType();

  constructor() {
    super();
  }

  toSource() {
    return 'date';
  }

  validate(value, {
    path = []
  } = {}) {
    return value instanceof Date ? [] : [{
      type: this,
      value,
      path
    }];
  }

  restrict(value, options = {}) {
    if (value instanceof Date) {
      return value;
    }

    if (options.coerce && (typeof value === 'string' || typeof value === 'number')) {
      try {
        return new Date(value);
      } catch (e) {// do nothing
      }
    }

    if (options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return new Date(0);
  }

}

exports.DateType = DateType;

class BooleanType extends ScalarType {
  static self = new BooleanType();

  constructor() {
    super();
  }

  toSource() {
    return 'boolean';
  }

  validate(value, {
    path = []
  } = {}) {
    return typeof value === 'boolean' ? [] : [{
      type: this,
      value,
      path
    }];
  }

  restrict(value, options = {}) {
    if (typeof value === 'boolean') {
      return value;
    } else if (options.coerce) {
      return !!value;
    } else if (options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return false;
  }

}

exports.BooleanType = BooleanType;

class NullType extends ScalarType {
  static self = new NullType();

  constructor() {
    super();
  }

  toSource() {
    return null;
  }

  source = null;

  validate(value, {
    path = []
  } = {}) {
    return value === null ? [] : [{
      type: this,
      value,
      path
    }];
  }

  restrict(value, options = {}) {
    if (value === null || options.coerce && !value || options.fillZero) {
      return null;
    }

    return undefined;
  }

  zeroValue() {
    return null;
  }

}

exports.NullType = NullType;

class OneOfType extends ScalarType {
  constructor(subtypes) {
    super();
    this.subtypes = subtypes;
  }

  toSource() {
    return ['oneof', ...this.subtypes.map(s => s.toSource())];
  }

  validate(value, options = {}) {
    var _options$path;

    return this.subtypes.some(s => !s.validate(value, options).length) ? [] : [{
      type: this,
      value,
      path: (_options$path = options.path) !== null && _options$path !== void 0 ? _options$path : []
    }];
  }

  restrict(value, options = {}) {
    for (const t of this.subtypes) {
      const restricted = t.restrict(value, {});

      if (restricted !== undefined) {
        return restricted;
      }
    }

    if (options.coerce) {
      for (const t of this.subtypes) {
        const coerced = t.restrict(value, {
          coerce: true
        });

        if (coerced !== undefined) {
          return coerced;
        }
      }
    }

    if (options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return this.subtypes[0].zeroValue();
  }

}

exports.OneOfType = OneOfType;

class EnumType extends ScalarType {
  constructor(members) {
    super();
    this.members = members;
  }

  toSource() {
    return ['enum', ...this.members];
  }

  validate(value, {
    path = []
  } = {}) {
    return this.members.indexOf(value) >= 0 ? [] : [{
      type: this,
      value,
      path
    }];
  }

  restrict(value, options = {}) {
    if (this.members.includes(value)) {
      return value;
    } else if (options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return this.members[0];
  }

}

exports.EnumType = EnumType;

class ArrayType extends Schema {
  constructor(elementType) {
    super();
    this.elementType = elementType;
  }

  toSource() {
    return [this.elementType.toSource()];
  }

  validate(value, options = {}) {
    var _options$path2;

    const path = (_options$path2 = options.path) !== null && _options$path2 !== void 0 ? _options$path2 : [];

    if (!Array.isArray(value)) {
      return [{
        type: this,
        value,
        path
      }];
    }

    return (0, _lodash.default)(value, (v, i) => this.elementType.validate(v, { ...options,
      path: [...path, i]
    }));
  }

  child(path) {
    return typeof path === 'number' ? this.elementType : undefined;
  }

  isArray() {
    return true;
  }

  isRecursive() {
    return this.elementType.isRecursive();
  }

  restrict(value, options = {}) {
    if (Array.isArray(value)) {
      return value.map(e => this.elementType.restrict(e, options)).filter(e => e !== undefined);
    } else if (options.fillEmpty || options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return [];
  }

}

exports.ArrayType = ArrayType;

class DictionaryType extends Schema {
  constructor(valueType) {
    super();
    this.valueType = valueType;
  }

  toSource() {
    return ['dictionary', this.valueType.toSource()];
  }

  validate(value, options = {}) {
    var _options$path3;

    const path = (_options$path3 = options.path) !== null && _options$path3 !== void 0 ? _options$path3 : [];

    if (!isObject(value)) {
      return [{
        type: this,
        value,
        path
      }];
    }

    return (0, _lodash.default)(Object.entries(value), ([k, v]) => this.valueType.validate(v, { ...options,
      path: [...path, k]
    }));
  }

  child(path) {
    return typeof path === 'string' ? this.valueType : undefined;
  }

  isObject() {
    return true;
  }

  isRecursive() {
    return this.valueType.isRecursive();
  }

  restrict(value, options = {}) {
    if (isObject(value)) {
      return (0, _lodash2.default)(Object.entries(value).map(([k, v]) => [k, this.valueType.restrict(v, options)]).filter(([, v]) => v !== undefined));
    } else if (options.fillEmpty || options.fillZero) {
      return this.zeroValue();
    }

    return undefined;
  }

  zeroValue() {
    return {};
  }

}

exports.DictionaryType = DictionaryType;

class ObjectType extends Schema {
  constructor(entries) {
    super();
    this.entries = entries;
  }

  toSource() {
    return (0, _lodash2.default)(Object.entries(this.entries).map(([k, v]) => [k, v.toSource()]));
  }

  validate(value, options = {}) {
    var _options$path4;

    const path = (_options$path4 = options.path) !== null && _options$path4 !== void 0 ? _options$path4 : [];

    if (!isObject(value)) {
      return [{
        type: this,
        value,
        path
      }];
    }

    const failures = (0, _lodash.default)(Object.entries(this.entries), ([k, v]) => k in value ? v.validate(value[k], { ...options,
      path: [...path, k]
    }) : []);

    if (!options.allowExtraFields && !Object.keys(value).every(k => k in this.entries)) {
      failures.push({
        type: this,
        value,
        path
      });
    }

    return failures;
  }

  child(path) {
    return this.entries[path];
  }

  isObject() {
    return true;
  }

  isRecursive() {
    return Object.values(this.entries).some(e => e.isRecursive());
  }

  restrict(value, options = {}) {
    if (!isObject(value) && !options.fillZero && !options.fillEmpty) {
      return undefined;
    }

    const base = isObject(value) ? value : {};
    return (0, _lodash2.default)(Object.entries(this.entries).map(([k, v]) => [k, v.restrict(base[k], options)]).filter(([, v]) => v !== undefined));
  }

  zeroValue() {
    return (0, _lodash2.default)(Object.entries(this.entries).map(([k, v]) => [k, v.zeroValue()]));
  }

}

exports.ObjectType = ObjectType;

class FixpointType extends Schema {
  constructor(label) {
    super();
    this.label = label;
  }

  toSource() {
    if (this.schema) {
      return this.schema.toSource();
    } else {
      return this.label;
    }
  }

  validate(value, options) {
    (0, _assert.default)(this.schema);
    return this.schema.validate(value, options);
  }

  child(path) {
    (0, _assert.default)(this.schema);
    return this.schema.child(path);
  }

  isArray() {
    (0, _assert.default)(this.schema);
    return this.schema.isArray();
  }

  isObject() {
    (0, _assert.default)(this.schema);
    return this.schema.isObject();
  }

  isRecursive() {
    return true;
  }

  restrict(value, options = {}) {
    if (options.fillEmpty || options.fillZero) {
      throw new _errors.RecursionError('cannot fillEmpty or fillZero on a recursive schema', this.label);
    }

    (0, _assert.default)(this.schema);
    return this.schema.restrict(value, options);
  }

  zeroValue() {
    throw new _errors.RecursionError('cannot compute zeroValue of a recursive schema', this.label);
  }

}

function compileSchema(source, path = [], scope = {}) {
  if (isObject(source)) {
    const [bindings, entries] = (0, _lodash4.default)(Object.entries(source), ([k]) => k.startsWith('$') && !k.startsWith('$$'));

    if (bindings.length) {
      if (typeof source['$schema'] === 'string' && source['$schema'].startsWith('http')) {
        throw new _errors.SchemaCompileError('this is a JSON Schema, not a Spartan Schema.', path);
      } // FIXME: Fixpoints pointing to scalars will break "oneof"


      const fixpoints = bindings.map(([k]) => [k, new FixpointType(k)]);
      const scopeWithFixpoints = { ...scope,
        ...(0, _lodash2.default)(fixpoints)
      };
      const compiledBindings = bindings.map(([k, v], i) => {
        const result = compileSchema(v, [...path, k], scopeWithFixpoints);
        fixpoints[i][1].schema = result;
        return [k, result];
      });
      scope = { ...scope,
        ...(0, _lodash2.default)(compiledBindings)
      };
    }

    return new ObjectType((0, _lodash2.default)(entries.map(([k, v]) => [k.startsWith('$$') ? k.slice(1) : k, compileSchema(v, [...path, k], scope)])));
  } else if (Array.isArray(source)) {
    if (source.length === 1) {
      return new ArrayType(compileSchema(source[0], [...path, 0], scope));
    } else if (source.length > 1) {
      switch (source[0]) {
        case 'oneof':
          return new OneOfType(source.slice(1).map((s, i) => {
            const result = compileSchema(s, [...path, i + 1], scope);

            if (!result.isScalar()) {
              throw new _errors.SchemaCompileError('oneof entries must be scalar types', [...path, i + 1]);
            }

            return result;
          }));

        case 'enum':
          return new EnumType(source.slice(1).map((s, i) => {
            switch (typeof s) {
              case 'string':
              case 'number':
              case 'boolean':
                return s;

              default:
                if (s === null) {
                  return s;
                }

                throw new _errors.SchemaCompileError('enum entries must be scalar values', [...path, i + 1]);
            }
          }));

        case 'dictionary':
          if (source.length !== 2) {
            throw new _errors.SchemaCompileError('dictionary must be a 2-element array', path);
          }

          return new DictionaryType(compileSchema(source[1], [...path, 1], scope));

        default:
          throw new _errors.SchemaCompileError(`not a Spartan Schema type tag: ${JSON.stringify(source[0])} (expected "oneof", "enum", or "dictionary")`, path);
      }
    }
  }

  switch (source) {
    case 'string':
      return StringType.self;

    case 'integer':
      return IntegerType.self;

    case 'float':
      return FloatType.self;

    case 'binary':
      return BinaryType.self;

    case 'date':
      return DateType.self;

    case 'boolean':
      return BooleanType.self;

    case 'null':
      return NullType.self;
  }

  if (typeof source === 'string') {
    if (source.startsWith('$') && !source.startsWith('$$')) {
      const ref = scope[source];

      if (ref) {
        return ref;
      }

      throw new _errors.SchemaCompileError(`label not in scope: ${JSON.stringify(source)}`, path);
    }

    throw new _errors.SchemaCompileError(`not a Spartan Schema type: ${JSON.stringify(source)}`, path);
  }

  if (source === null) {
    return NullType.self;
  }

  throw new _errors.SchemaCompileError(`unexpected JSON value: ${JSON.stringify(source)}`, path);
}